# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_config import cfg
from oslo_messaging.rpc import dispatcher as rpc
from oslo_utils import uuidutils
import six

from senlin.common import consts
from senlin.common import exception as exc
from senlin.common.i18n import _
from senlin.common import scaleutils as su
from senlin.common import utils as common_utils
from senlin.engine.actions import base as am
from senlin.engine import cluster as cm
from senlin.engine import dispatcher
from senlin.engine import node as nm
from senlin.engine import service
from senlin.objects import cluster as co
from senlin.objects import cluster_policy as cpo
from senlin.objects import receiver as ro
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class ClusterTest(base.SenlinTestCase):

    def setUp(self):
        super(ClusterTest, self).setUp()

        self.ctx = utils.dummy_context(project='cluster_test_project')
        self.eng = service.EngineService('host-a', 'topic-a')

    @mock.patch.object(co.Cluster, 'get')
    def test_cluster_find_by_uuid(self, mock_get):
        x_cluster = mock.Mock()
        mock_get.return_value = x_cluster

        aid = uuidutils.generate_uuid()
        result = self.eng.cluster_find(self.ctx, aid)

        self.assertEqual(x_cluster, result)
        mock_get.assert_called_once_with(self.ctx, aid, project_safe=True)

    @mock.patch.object(co.Cluster, 'get_by_name')
    @mock.patch.object(co.Cluster, 'get')
    def test_cluster_find_by_uuid_as_name(self, mock_get, mock_get_name):
        x_cluster = mock.Mock()
        mock_get_name.return_value = x_cluster
        mock_get.return_value = None

        aid = uuidutils.generate_uuid()
        result = self.eng.cluster_find(self.ctx, aid, False)

        self.assertEqual(x_cluster, result)
        mock_get.assert_called_once_with(self.ctx, aid, project_safe=False)
        mock_get_name.assert_called_once_with(self.ctx, aid,
                                              project_safe=False)

    @mock.patch.object(co.Cluster, 'get_by_name')
    def test_cluster_find_by_name(self, mock_get_name):
        x_cluster = mock.Mock()
        mock_get_name.return_value = x_cluster

        aid = 'this-is-not-uuid'
        result = self.eng.cluster_find(self.ctx, aid)

        self.assertEqual(x_cluster, result)
        mock_get_name.assert_called_once_with(self.ctx, aid, project_safe=True)

    @mock.patch.object(co.Cluster, 'get_by_short_id')
    @mock.patch.object(co.Cluster, 'get_by_name')
    def test_cluster_find_by_shortid(self, mock_get_name, mock_get_shortid):
        x_cluster = mock.Mock()
        mock_get_shortid.return_value = x_cluster
        mock_get_name.return_value = None

        aid = 'abcd-1234-abcd'
        result = self.eng.cluster_find(self.ctx, aid, False)

        self.assertEqual(x_cluster, result)
        mock_get_name.assert_called_once_with(self.ctx, aid,
                                              project_safe=False)
        mock_get_shortid.assert_called_once_with(self.ctx, aid,
                                                 project_safe=False)

    @mock.patch.object(co.Cluster, 'get_by_name')
    def test_cluster_find_not_found(self, mock_get_name):
        mock_get_name.return_value = None

        ex = self.assertRaises(exc.ResourceNotFound,
                               self.eng.cluster_find,
                               self.ctx, 'bogus')

        self.assertEqual('The cluster (bogus) could not be found.',
                         six.text_type(ex))
        mock_get_name.assert_called_once_with(self.ctx, 'bogus',
                                              project_safe=True)

    @mock.patch.object(cm.Cluster, 'load_all')
    def test_cluster_list(self, mock_load):
        x_obj_1 = mock.Mock()
        x_obj_1.to_dict.return_value = {'k': 'v1'}
        x_obj_2 = mock.Mock()
        x_obj_2.to_dict.return_value = {'k': 'v2'}
        mock_load.return_value = [x_obj_1, x_obj_2]

        result = self.eng.cluster_list(self.ctx)

        self.assertEqual([{'k': 'v1'}, {'k': 'v2'}], result)
        mock_load.assert_called_once_with(self.ctx, limit=None, marker=None,
                                          filters=None, sort=None,
                                          project_safe=True)

    @mock.patch.object(cm.Cluster, 'load_all')
    def test_cluster_list_with_params(self, mock_load):
        mock_load.return_value = []

        result = self.eng.cluster_list(self.ctx, limit=10, marker='KEY',
                                       filters={'foo': 'bar'}, sort='name:asc',
                                       project_safe=True)

        self.assertEqual([], result)
        mock_load.assert_called_once_with(self.ctx, limit=10, marker='KEY',
                                          filters={'foo': 'bar'},
                                          sort='name:asc',
                                          project_safe=True)

    def test_cluster_list_bad_param(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_list,
                               self.ctx, limit='no')
        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_list,
                               self.ctx, sort='crazykey')
        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_list,
                               self.ctx, project_safe='no')
        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])

    @mock.patch.object(cm.Cluster, 'load_all')
    def test_cluster_list_with_project_safe(self, mock_load):
        mock_load.return_value = []

        result = self.eng.cluster_list(self.ctx)
        self.assertEqual([], result)
        mock_load.assert_called_once_with(self.ctx, filters=None, limit=None,
                                          sort=None, marker=None,
                                          project_safe=True)
        mock_load.reset_mock()

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_list,
                               self.ctx, project_safe=False)
        self.assertEqual(exc.Forbidden, ex.exc_info[0])

        self.ctx.is_admin = True

        result = self.eng.cluster_list(self.ctx)
        self.assertEqual([], result)
        mock_load.assert_called_once_with(self.ctx, filters=None, limit=None,
                                          sort=None, marker=None,
                                          project_safe=True)
        mock_load.reset_mock()

        result = self.eng.cluster_list(self.ctx, project_safe=True)
        self.assertEqual([], result)
        mock_load.assert_called_once_with(self.ctx, filters=None, limit=None,
                                          sort=None, marker=None,
                                          project_safe=True)
        mock_load.reset_mock()

        result = self.eng.cluster_list(self.ctx, project_safe=False)
        self.assertEqual([], result)
        mock_load.assert_called_once_with(self.ctx, filters=None, limit=None,
                                          sort=None, marker=None,
                                          project_safe=False)

    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_get(self, mock_find, mock_load):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        x_cluster = mock.Mock()
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_load.return_value = x_cluster

        result = self.eng.cluster_get(self.ctx, 'CLUSTER')

        self.assertEqual({'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_load.assert_called_once_with(self.ctx, dbcluster=x_obj)

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_get_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_get,
                               self.ctx, 'Bogus')
        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(am.Action, 'create')
    @mock.patch("senlin.engine.cluster.Cluster")
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_create(self, notify, mock_profile, mock_cluster,
                            mock_action, mock_check, mock_quota):
        x_profile = mock.Mock(id='PROFILE_ID')
        mock_profile.return_value = x_profile
        x_cluster = mock.Mock(id='12345678ABC')
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_cluster.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'
        mock_check.return_value = None
        mock_quota.return_value = None

        result = self.eng.cluster_create(self.ctx, 'C1', 3, 'PROFILE')

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_profile.assert_called_once_with(self.ctx, 'PROFILE')
        mock_check.assert_called_once_with(None, 3, None, None, True)
        mock_cluster.assert_called_once_with(
            'C1', 3, 'PROFILE_ID',
            min_size=None, max_size=None, timeout=3600, metadata=None,
            user=self.ctx.user, project=self.ctx.project,
            domain=self.ctx.domain)
        x_cluster.store.assert_called_once_with(self.ctx)
        mock_action.assert_called_once_with(
            self.ctx,
            '12345678ABC', 'CLUSTER_CREATE',
            name='cluster_create_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
        )
        notify.assert_called_once_with()

    @mock.patch.object(co.Cluster, 'count_all')
    def test_check_cluster_quota(self, mock_count):
        mock_count.return_value = 10
        cfg.CONF.set_override('max_clusters_per_project', 11,
                              enforce_type=True)

        res = self.eng.check_cluster_quota(self.ctx)

        self.assertIsNone(res)
        mock_count.assert_called_once_with(self.ctx)

    @mock.patch.object(co.Cluster, 'count_all')
    def test_check_cluster_quota_failed(self, mock_count):
        mock_count.return_value = 11
        cfg.CONF.set_override('max_clusters_per_project', 11,
                              enforce_type=True)

        ex = self.assertRaises(exc.Forbidden,
                               self.eng.check_cluster_quota, self.ctx)
        self.assertEqual("You are not authorized to complete this "
                         "operation.",
                         six.text_type(ex))

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    def test_cluster_create_exceeding_quota(self, mock_quota):
        mock_quota.side_effect = exc.Forbidden()

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, 'CLUSTER', 123, 'PROFILE')

        self.assertEqual(exc.Forbidden, ex.exc_info[0])
        self.assertEqual("You are not authorized to complete this "
                         "operation.",
                         six.text_type(ex.exc_info[1]))
        mock_quota.assert_called_once_with(self.ctx)

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(co.Cluster, 'get_by_name')
    def test_cluster_create_already_exists(self, mock_get, mock_quota):
        mock_quota.return_value = None
        cfg.CONF.set_override('name_unique', True, enforce_type=True)

        x_cluster = mock.Mock()
        mock_get.return_value = x_cluster

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, 'CLUSTER', 123, 'PROFILE')

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual(_("The request is malformed: The cluster (CLUSTER) "
                           "already exists."),
                         six.text_type(ex.exc_info[1]))
        mock_get.assert_called_once_with(self.ctx, 'CLUSTER')

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(service.EngineService, 'profile_find')
    def test_cluster_create_profile_not_found(self, mock_find, mock_quota):
        mock_quota.return_value = None
        mock_find.side_effect = exc.ResourceNotFound(type='profile',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, 'CLUSTER', 2, 'Bogus')

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: "
                         "The specified profile (Bogus) could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(service.EngineService, 'profile_find')
    def test_cluster_create_bad_desired_capacity(self, mock_find, mock_quota):
        mock_quota.return_value = None
        mock_find.return_value = mock.Mock()

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, 'CLUSTER', -2, 'PROFILE')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value '-2' specified for 'desired_capacity'",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'PROFILE')

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(service.EngineService, 'profile_find')
    def test_cluster_create_bad_min_size(self, mock_find, mock_quota):
        mock_quota.return_value = None
        mock_find.return_value = mock.Mock()

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, 'CLUSTER', 2, 'PROFILE', min_size=-1)

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value '-1' specified for 'min_size'",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'PROFILE')

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(service.EngineService, 'profile_find')
    def test_cluster_create_bad_max_size(self, mock_find, mock_quota):
        mock_quota.return_value = None
        mock_find.return_value = mock.Mock()

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, 'CLUSTER', 2, 'PROFILE',
                               max_size='BIG')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'BIG' specified for 'max_size'",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'PROFILE')

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(service.EngineService, 'profile_find')
    def test_cluster_create_bad_timeout(self, mock_find, mock_quota):
        mock_quota.return_value = None
        mock_find.return_value = mock.Mock()

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, 'CLUSTER', 2, 'PROFILE', timeout=-1)

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value '-1' specified for 'timeout'",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'PROFILE')

    @mock.patch.object(service.EngineService, 'check_cluster_quota')
    @mock.patch.object(service.EngineService, 'profile_find')
    def test_cluster_create_failed_checking(self, mock_find, mock_quota):
        mock_quota.return_value = None
        mock_find.return_value = mock.Mock()

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_create,
                               self.ctx, 'CLUSTER', 2, 'PROFILE',
                               min_size=3)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: The target capacity (2) "
                         "is less than the specified min_size (3).",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'PROFILE')

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_update(self, notify, mock_find, mock_profile,
                            mock_load, mock_action):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              profile_id='OLD_PROFILE',
                              metadata={'A': 'B'})
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_load.return_value = x_cluster
        old_profile = mock.Mock(type='FAKE_TYPE', id='ID_OLD')
        new_profile = mock.Mock(type='FAKE_TYPE', id='ID_NEW')
        mock_profile.side_effect = [old_profile, new_profile]
        mock_action.return_value = 'ACTION_ID'

        # do it
        result = self.eng.cluster_update(self.ctx, 'FAKE_ID', name='new_name',
                                         metadata={'B': 'A'}, timeout=120,
                                         profile_id='NEW_PROFILE')

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_load.assert_called_once_with(self.ctx, dbcluster=x_obj)
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'OLD_PROFILE'),
            mock.call(self.ctx, 'NEW_PROFILE'),
        ])
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_UPDATE',
            name='cluster_update_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                'new_profile_id': 'ID_NEW',
                'metadata': {
                    'B': 'A',
                },
                'timeout': 120,
                'name': 'new_name',
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_update_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, 'Bogus')
        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])

    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_update_cluster_bad_status(self, mock_find, mock_load):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        mock_load.return_value = mock.Mock(status='ERROR', ERROR='ERROR')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, 'CLUSTER', name='new_name')

        self.assertEqual(exc.FeatureNotSupported, ex.exc_info[0])
        self.assertEqual('Updating a cluster in error state is not supported.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_load.assert_called_once_with(self.ctx, dbcluster=x_obj)

    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_update_profile_not_found(self, mock_find, mock_load,
                                              mock_profile):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        mock_load.return_value = mock.Mock(status='ACTIVE', ERROR='ERROR',
                                           profile_id='OLD_ID')

        mock_profile.side_effect = [
            mock.Mock(type='FAKE_TYPE', id='OLD_ID'),
            exc.ResourceNotFound(type='profile', id='Bogus')
        ]

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, 'CLUSTER', profile_id='Bogus')

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: "
                         "The specified profile (Bogus) could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_load.assert_called_once_with(self.ctx, dbcluster=x_obj)
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'OLD_ID'),
            mock.call(self.ctx, 'Bogus'),
        ])

    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_update_diff_profile_type(self, mock_find, mock_load,
                                              mock_profile):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        mock_load.return_value = mock.Mock(status='ACTIVE', ERROR='ERROR',
                                           profile_id='OLD_ID')

        mock_profile.side_effect = [
            mock.Mock(type='FAKE_TYPE', id='OLD_ID'),
            mock.Mock(type='DIFF_TYPE', id='NEW_ID'),
        ]

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, 'CLUSTER', profile_id='NEW_PROFILE')

        self.assertEqual(exc.ProfileTypeNotMatch, ex.exc_info[0])
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_load.assert_called_once_with(self.ctx, dbcluster=x_obj)
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'OLD_ID'),
            mock.call(self.ctx, 'NEW_PROFILE'),
        ])

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_update_same_profile(self, notify, mock_find, mock_profile,
                                         mock_load, mock_action):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              profile_id='OLD_PROFILE')
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_load.return_value = x_cluster
        old_profile = mock.Mock(type='FAKE_TYPE', id='ID_OLD')
        new_profile = mock.Mock(type='FAKE_TYPE', id='ID_OLD')
        mock_profile.side_effect = [old_profile, new_profile]
        mock_action.return_value = 'ACTION_ID'

        # do it
        result = self.eng.cluster_update(self.ctx, 'FAKE_ID', name='new_name',
                                         profile_id='NEW_PROFILE')

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_load.assert_called_once_with(self.ctx, dbcluster=x_obj)
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'OLD_PROFILE'),
            mock.call(self.ctx, 'NEW_PROFILE'),
        ])
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_UPDATE',
            name='cluster_update_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                # Note profile_id is not shown in the inputs
                'name': 'new_name',
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_update_same_metadata(self, notify, mock_find, mock_load,
                                          mock_action):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE',
                              metadata={'K': 'V'})
        x_cluster.to_dict.return_value = {'foo': 'bar'}
        mock_load.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'

        # do it
        result = self.eng.cluster_update(self.ctx, 'FAKE_ID', name='new_name',
                                         metadata={'K': 'V'})

        self.assertEqual({'action': 'ACTION_ID', 'foo': 'bar'}, result)
        mock_find.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_load.assert_called_once_with(self.ctx, dbcluster=x_obj)
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_UPDATE',
            name='cluster_update_12345678',
            status=am.Action.READY,
            cause=am.CAUSE_RPC,
            inputs={
                # Note metadata is not included in the inputs
                'name': 'new_name',
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_update_timeout_not_integer(self, mock_find, mock_load):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        x_cluster = mock.Mock(id='12345678AB', status='ACTIVE')
        mock_load.return_value = x_cluster

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, 'CLUSTER', timeout='Long')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_load.assert_called_once_with(self.ctx, dbcluster=x_obj)

    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_update_cluster_no_property_updated(self, mock_find,
                                                        mock_load):
        x_obj = mock.Mock()
        mock_find.return_value = x_obj
        mock_load.return_value = mock.Mock(status='ACTIVE', ERROR='ERROR',
                                           profile_id='OLD_ID')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_update,
                               self.ctx, 'CLUSTER')
        self.assertEqual(exc.BadRequest, ex.exc_info[0])

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(ro.Receiver, 'get_all')
    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_delete(self, notify, mock_find, mock_policies,
                            mock_receivers, mock_action):
        x_obj = mock.Mock(id='12345678AB', status='ACTIVE', dependents={})
        mock_find.return_value = x_obj
        mock_policies.return_value = []
        mock_receivers.return_value = []
        mock_action.return_value = 'ACTION_ID'

        result = self.eng.cluster_delete(self.ctx, 'IDENTITY')

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'IDENTITY')
        mock_policies.assert_called_once_with(self.ctx, '12345678AB')
        mock_receivers.assert_called_once_with(
            self.ctx, filters={'cluster_id': '12345678AB'})
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', 'CLUSTER_DELETE',
            name='cluster_delete_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY)

        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_delete_contain_container(self, mock_find):
        dependents = {'containers': ['container1']}
        cluster = mock.Mock(id='cluster1', status='ACTIVE',
                            dependents=dependents)
        mock_find.return_value = cluster
        identity = mock.Mock()
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, identity)
        msg = _('The host_cluster (cluster1) is still in use.')
        self.assertEqual(exc.ResourceInUse, ex.exc_info[0])
        self.assertEqual(msg, six.text_type(ex.exc_info[1]))

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_delete_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, 'Bogus')

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual('The cluster (Bogus) could not be found.',
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_node_delete_improper_status(self, mock_find):
        for bad_status in [cm.Cluster.CREATING, cm.Cluster.UPDATING,
                           cm.Cluster.DELETING, cm.Cluster.RECOVERING]:
            fake_cluster = mock.Mock(id='12345678AB', status=bad_status)
            mock_find.return_value = fake_cluster
            ex = self.assertRaises(rpc.ExpectedException,
                                   self.eng.cluster_delete,
                                   self.ctx, 'BUSY')

            self.assertEqual(exc.ActionInProgress, ex.exc_info[0])
            self.assertEqual("The cluster BUSY is in status %s." % bad_status,
                             six.text_type(ex.exc_info[1]))
            # skipping assertion on mock_find

    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_delete_policy_attached(self, mock_find, mock_policies):
        x_obj = mock.Mock(id='12345678AB', dependents={})
        mock_find.return_value = x_obj
        mock_policies.return_value = [mock.Mock()]

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, 'IDENTITY')

        self.assertEqual(exc.ClusterBusy, ex.exc_info[0])
        expected_msg = _('The cluster (IDENTITY) cannot be deleted: '
                         'there is still policy(s) attached to it.')
        self.assertEqual(expected_msg, six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'IDENTITY')
        mock_policies.assert_called_once_with(self.ctx, '12345678AB')

    @mock.patch.object(ro.Receiver, 'get_all')
    @mock.patch.object(cpo.ClusterPolicy, 'get_all')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_delete_with_receiver(self, mock_find, mock_policies,
                                          mock_receivers):
        x_obj = mock.Mock(id='12345678AB', dependents={})
        mock_find.return_value = x_obj
        mock_policies.return_value = []
        mock_receivers.return_value = [mock.Mock()]

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_delete,
                               self.ctx, 'IDENTITY')

        self.assertEqual(exc.ClusterBusy, ex.exc_info[0])
        expected_msg = _('The cluster (IDENTITY) cannot be deleted: '
                         'there is still receiver(s) associated with it.')
        self.assertEqual(expected_msg, six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'IDENTITY')
        mock_policies.assert_called_once_with(self.ctx, '12345678AB')
        mock_receivers.assert_called_once_with(
            self.ctx, filters={'cluster_id': '12345678AB'})

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_add_nodes(self, notify, mock_find, mock_profile,
                               mock_node, mock_action, mock_check):
        x_cluster = mock.Mock(id='12345678AB', profile_id='FAKE_ID',
                              desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        x_node_1 = mock.Mock(id='NODE1', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_ID_1')
        x_node_2 = mock.Mock(id='NODE2', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_ID_1')
        mock_node.side_effect = [x_node_1, x_node_2]
        mock_action.return_value = 'ACTION_ID'
        mock_check.return_value = None

        result = self.eng.cluster_add_nodes(self.ctx, 'C1',
                                            ['NODE_A', 'NODE_B'])

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'NODE_A'),
            mock.call(self.ctx, 'NODE_B'),
        ])
        mock_check.assert_called_once_with(x_cluster, 6, strict=True)
        mock_action.assert_called_once_with(
            self.ctx, '12345678AB', consts.CLUSTER_ADD_NODES,
            name='cluster_add_nodes_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'nodes': ['NODE1', 'NODE2']},
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_add_nodes_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, 'Bogus', ['n1', 'n2'])

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual('The cluster (Bogus) could not be found.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_add_nodes_empty_list(self, mock_find, mock_profile):

        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, 'CLUSTER', [])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual('The request is malformed: No nodes to add: [].',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_called_once_with(self.ctx, 'FAKE_ID')

    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_add_nodes_nodes_not_found(self, mock_find, mock_profile,
                                               mock_node):
        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        mock_node.side_effect = exc.ResourceNotFound(type='node', id='NODE1')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, 'CLUSTER', ['NODE1'])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Nodes not found: "
                         "['NODE1'].", six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_node.assert_called_once_with(self.ctx, 'NODE1')

    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_add_nodes_bad_status(self, mock_find, mock_profile,
                                          mock_node):
        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        mock_node.return_value = mock.Mock(id='NODE2', status='ERROR')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, 'CLUSTER', ['NODE2'])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Nodes are not ACTIVE: "
                         "['NODE2'].",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_node.assert_called_once_with(self.ctx, 'NODE2')

    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_add_nodes_node_already_owned(self, mock_find,
                                                  mock_profile, mock_node):

        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        mock_node.return_value = mock.Mock(id='NODE3', status='ACTIVE',
                                           cluster_id='OTHER')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, 'CLUSTER', ['NODE3'])

        self.assertEqual(exc.NodeNotOrphan, ex.exc_info[0])
        self.assertEqual("Nodes ['NODE3'] already owned by some cluster.",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_called_once_with(self.ctx, 'FAKE_ID')
        mock_node.assert_called_once_with(self.ctx, 'NODE3')

    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_add_nodes_node_profile_type_not_match(
            self, mock_find, mock_profile, mock_node):

        mock_find.return_value = mock.Mock(id='1234', profile_id='FAKE_ID')
        mock_profile.side_effect = [
            mock.Mock(type='FAKE_TYPE_1'),
            mock.Mock(type='FAKE_TYPE_2'),
        ]
        mock_node.return_value = mock.Mock(id='NODE4', status='ACTIVE',
                                           cluster_id='', profile_id='DIFF')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, 'CLUSTER', ['NODE4'])

        self.assertEqual(exc.ProfileTypeNotMatch, ex.exc_info[0])
        self.assertEqual("Profile type of nodes ['NODE4'] does not match "
                         "that of the cluster.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'FAKE_ID'),
            mock.call(self.ctx, 'DIFF'),
        ])
        mock_node.assert_called_once_with(self.ctx, 'NODE4')

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'profile_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_add_nodes_failed_checking(self, mock_find, mock_profile,
                                               mock_node, mock_check):
        x_cluster = mock.Mock(id='12345678AB', profile_id='FAKE_PROFILE',
                              desired_capacity=2)
        mock_find.return_value = x_cluster
        mock_profile.return_value = mock.Mock(type='FAKE_TYPE')
        x_node_1 = mock.Mock(id='NODE1', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_PROFILE_1')
        x_node_2 = mock.Mock(id='NODE2', cluster_id='', status='ACTIVE',
                             profile_id='FAKE_PROFILE_2')
        mock_node.side_effect = [x_node_1, x_node_2]
        mock_check.return_value = 'Failed size checking.'

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_add_nodes,
                               self.ctx, 'C1', ['NODE_A', 'NODE_B'])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Failed size checking.",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'C1')
        mock_profile.assert_has_calls([
            mock.call(self.ctx, 'FAKE_PROFILE'),
            mock.call(self.ctx, 'FAKE_PROFILE_1'),
            mock.call(self.ctx, 'FAKE_PROFILE_2'),
        ])
        mock_node.assert_has_calls([
            mock.call(self.ctx, 'NODE_A'),
            mock.call(self.ctx, 'NODE_B'),
        ])
        mock_check.assert_called_once_with(x_cluster, 4, strict=True)

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_del_nodes(self, notify, mock_find, mock_node,
                               mock_action, mock_check):
        x_cluster = mock.Mock(id='1234', desired_capacity=2)
        mock_find.return_value = x_cluster
        mock_node.return_value = mock.Mock(id='NODE2', cluster_id='1234')
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'

        result = self.eng.cluster_del_nodes(self.ctx, 'CLUSTER', ['NODE1'])

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE1')
        mock_check.asset_called_once_with(x_cluster, 1, strict=True)
        mock_action.assert_called_once_with(
            self.ctx, '1234', consts.CLUSTER_DEL_NODES,
            name='cluster_del_nodes_1234',
            status=am.Action.READY,
            cause=am.CAUSE_RPC,
            inputs={
                'count': 1,
                'candidates': ['NODE2'],
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_del_nodes_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, 'Bogus', ['n1', 'n2'])

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual('The cluster (Bogus) could not be found.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_del_nodes_empty_list(self, mock_find):
        mock_find.return_value = mock.Mock(id='1234')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, 'CLUSTER', [])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual('The request is malformed: No nodes specified.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')

    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_del_nodes_node_not_found(self, mock_find, mock_node):
        mock_find.return_value = mock.Mock()
        mock_node.side_effect = exc.ResourceNotFound(type='node', id='NODE1')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, 'CLUSTER', ['NODE1'])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Nodes not found: "
                         "['NODE1'].", six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE1')

    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_del_nodes_node_in_other_cluster(self, mock_find,
                                                     mock_node):
        mock_find.return_value = mock.Mock(id='1234')
        mock_node.return_value = mock.Mock(id='NODE2', cluster_id='5678')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, 'CLUSTER', ['NODE2'])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Nodes not members of "
                         "specified cluster: ['NODE2'].",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE2')

    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_del_nodes_orphan_nodes(self, mock_find, mock_node):
        mock_find.return_value = mock.Mock(id='1234')
        mock_node.return_value = mock.Mock(id='NODE3', cluster_id='')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, 'CLUSTER', ['NODE3'])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Nodes not members of "
                         "specified cluster: ['NODE3'].",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE3')

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(service.EngineService, 'node_find')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_del_nodes_failed_checking(self, mock_find, mock_node,
                                               mock_check):
        x_cluster = mock.Mock(id='1234', desired_capacity=2)
        mock_find.return_value = x_cluster
        mock_node.return_value = mock.Mock(id='NODE2', cluster_id='1234')
        mock_check.return_value = 'Failed size checking.'

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_del_nodes,
                               self.ctx, 'CLUSTER', ['NODE1'])

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Failed size checking.",
                         six.text_type(ex.exc_info[1]))

        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_node.assert_called_once_with(self.ctx, 'NODE1')
        mock_check.assert_called_once_with(x_cluster, 1, strict=True)

    @mock.patch.object(su, 'calculate_desired')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_resize_exact_capacity(self, mock_find, mock_action,
                                           notify, mock_check, mock_calc):
        x_cluster = mock.Mock(id='12345678ABCDEFGH', desired_capacity=3)
        mock_find.return_value = x_cluster
        mock_calc.return_value = 5
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'

        res = self.eng.cluster_resize(self.ctx, 'CLUSTER',
                                      adj_type=consts.EXACT_CAPACITY,
                                      number=5)

        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_calc.assert_called_once_with(3, consts.EXACT_CAPACITY, 5, None)
        mock_check.assert_called_once_with(x_cluster, 5, None, None, True)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_RESIZE,
            name='cluster_resize_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                consts.ADJUSTMENT_TYPE: consts.EXACT_CAPACITY,
                consts.ADJUSTMENT_NUMBER: 5,
                consts.ADJUSTMENT_MIN_SIZE: None,
                consts.ADJUSTMENT_MAX_SIZE: None,
                consts.ADJUSTMENT_MIN_STEP: None,
                consts.ADJUSTMENT_STRICT: True
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(su, 'calculate_desired')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_resize_change_in_capacity(self, mock_find, mock_action,
                                               notify, mock_check, mock_calc):
        x_cluster = mock.Mock(id='12345678ABCDEFGH', desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_calc.return_value = 9
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'

        res = self.eng.cluster_resize(self.ctx, 'CLUSTER',
                                      adj_type=consts.CHANGE_IN_CAPACITY,
                                      number=5)

        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_calc.assert_called_once_with(4, consts.CHANGE_IN_CAPACITY, 5,
                                          None)
        mock_check.assert_called_once_with(x_cluster, 9, None, None, True)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_RESIZE,
            name='cluster_resize_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                consts.ADJUSTMENT_TYPE: consts.CHANGE_IN_CAPACITY,
                consts.ADJUSTMENT_NUMBER: 5,
                consts.ADJUSTMENT_MIN_SIZE: None,
                consts.ADJUSTMENT_MAX_SIZE: None,
                consts.ADJUSTMENT_MIN_STEP: None,
                consts.ADJUSTMENT_STRICT: True
            },
        )
        notify.assert_called_once_with()

    @mock.patch.object(su, 'calculate_desired')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_resize_change_in_percentage(self, mock_find, mock_action,
                                                 notify, mock_check,
                                                 mock_calc):
        x_cluster = mock.Mock(id='12345678ABCDEFGH', desired_capacity=10)
        mock_find.return_value = x_cluster
        mock_calc.return_value = 8
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'

        res = self.eng.cluster_resize(self.ctx, 'CLUSTER',
                                      adj_type=consts.CHANGE_IN_PERCENTAGE,
                                      number=15.81)

        self.assertEqual({'action': 'ACTION_ID'}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_calc.assert_called_once_with(10, consts.CHANGE_IN_PERCENTAGE,
                                          15.81, None)
        mock_check.assert_called_once_with(x_cluster, 8, None, None, True)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_RESIZE,
            name='cluster_resize_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={
                consts.ADJUSTMENT_TYPE: consts.CHANGE_IN_PERCENTAGE,
                consts.ADJUSTMENT_NUMBER: 15.81,
                consts.ADJUSTMENT_MIN_SIZE: None,
                consts.ADJUSTMENT_MAX_SIZE: None,
                consts.ADJUSTMENT_MIN_STEP: None,
                consts.ADJUSTMENT_STRICT: True
            },
        )
        notify.assert_called_once_with()

    def test_cluster_resize_bad_adj_type(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type='BOOMBOOM')
        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'BOOMBOOM' specified for "
                         "'adjustment_type'", six.text_type(ex.exc_info[1]))

    def test_cluster_resize_missing_number(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type=consts.EXACT_CAPACITY)
        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Missing number value for "
                         "size adjustment.", six.text_type(ex.exc_info[1]))

    def test_cluster_resize_number_without_type(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER', number=10)
        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: Missing adjustment_type "
                         "value for size adjustment.",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_bad_number_for_exact_capacity(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type=consts.EXACT_CAPACITY,
                               number='BIGGER')
        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'BIGGER' specified for 'number'",
                         six.text_type(ex.exc_info[1]))

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type=consts.EXACT_CAPACITY,
                               number=-10)
        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value '-10' specified for 'number'",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_bad_number_for_change_capacity(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type=consts.CHANGE_IN_CAPACITY,
                               number='BIGGER')
        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'BIGGER' specified for 'number'",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_bad_number_for_change_percentage(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type=consts.CHANGE_IN_PERCENTAGE,
                               number='BIGGER')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'BIGGER' specified for 'number'",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_bad_step_for_change_percentage(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type=consts.CHANGE_IN_PERCENTAGE,
                               number=10.2, min_step='SMALL')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'SMALL' specified for 'min_step'",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_bad_min_size(self):
        # we intentionally omit adj_type and number here because we allow
        # the use case where only min_size/max_size is specified
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER', min_size='SMALL')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'SMALL' specified for 'min_size'",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_bad_max_size(self):
        # we intentionally omit adj_type and number here because we allow
        # the use case where only min_size/max_size is specified
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER', max_size='INF')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'INF' specified for 'max_size'",
                         six.text_type(ex.exc_info[1]))

    def test_cluster_resize_bad_strict(self):
        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER', strict='YES')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'YES' specified for 'strict'",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_resize_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='FAKE_CLUSTER')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type=consts.EXACT_CAPACITY, number=10)

        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual("The cluster (FAKE_CLUSTER) could not be found.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_resize_failing_size_check(self, mock_find, mock_check):
        mock_find.return_value = mock.Mock()
        mock_check.return_value = 'size check.'

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_resize,
                               self.ctx, 'FAKE_CLUSTER',
                               adj_type=consts.EXACT_CAPACITY, number=10)

        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: size check.",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_out(self, mock_find, mock_check, mock_action,
                               notify):
        x_cluster = mock.Mock(id='12345678ABCDEFGH', desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'

        result = self.eng.cluster_scale_out(self.ctx, 'CLUSTER', count=1)

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_check.assert_called_once_with(x_cluster, 5)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_SCALE_OUT,
            name='cluster_scale_out_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'count': 1},
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_out_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_out,
                               self.ctx, 'Bogus')

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual('The cluster (Bogus) could not be found.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_out_count_is_none(self, mock_find, mock_action,
                                             notify):
        mock_find.return_value = mock.Mock(id='12345678ABCDEFGH',
                                           desired_capacity=4)
        mock_action.return_value = 'ACTION_ID'

        result = self.eng.cluster_scale_out(self.ctx, 'CLUSTER')

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCDEFGH', consts.CLUSTER_SCALE_OUT,
            name='cluster_scale_out_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={},
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_out_count_not_int_or_zero(self, mock_find):
        mock_find.return_value = mock.Mock(desired_capacity=4)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_out,
                               self.ctx, 'CLUSTER', count='one')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'one' specified for 'count'",
                         six.text_type(ex.exc_info[1]))

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_out,
                               self.ctx, 'CLUSTER', count=0)

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value '0' specified for 'count'",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_out_failed_size_check(self, mock_find, mock_check):
        x_cluster = mock.Mock(desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_check.return_value = 'size limit'

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_out,
                               self.ctx, 'FAKE_CLUSTER', 2)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: size limit.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_check.assert_called_once_with(x_cluster, 6)

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_in(self, mock_find, mock_check, mock_action,
                              notify):
        x_cluster = mock.Mock(id='12345678ABCD', desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_check.return_value = None
        mock_action.return_value = 'ACTION_ID'

        result = self.eng.cluster_scale_in(self.ctx, 'CLUSTER', count=1)

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_check.assert_called_once_with(x_cluster, 3)
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCD', consts.CLUSTER_SCALE_IN,
            name='cluster_scale_in_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'count': 1},
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_in_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_in,
                               self.ctx, 'Bogus')

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual('The cluster (Bogus) could not be found.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_in_count_is_none(self, mock_find, mock_action,
                                            notify):
        mock_find.return_value = mock.Mock(id='12345678ABCD',
                                           desired_capacity=4)
        mock_action.return_value = 'ACTION_ID'

        result = self.eng.cluster_scale_in(self.ctx, 'CLUSTER')

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_action.assert_called_once_with(
            self.ctx, '12345678ABCD', consts.CLUSTER_SCALE_IN,
            name='cluster_scale_in_12345678',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={},
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_in_count_not_int_or_zero(self, mock_find):
        mock_find.return_value = mock.Mock(desired_capacity=4)

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_in,
                               self.ctx, 'CLUSTER', count='one')

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value 'one' specified for 'count'",
                         six.text_type(ex.exc_info[1]))

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_in,
                               self.ctx, 'CLUSTER', count=0)

        self.assertEqual(exc.InvalidParameter, ex.exc_info[0])
        self.assertEqual("Invalid value '0' specified for 'count'",
                         six.text_type(ex.exc_info[1]))

    @mock.patch.object(su, 'check_size_params')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_scale_in_failed_size_check(self, mock_find, mock_check):
        x_cluster = mock.Mock(desired_capacity=4)
        mock_find.return_value = x_cluster
        mock_check.return_value = 'size limit'

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_scale_in,
                               self.ctx, 'FAKE_CLUSTER', 2)

        self.assertEqual(exc.BadRequest, ex.exc_info[0])
        self.assertEqual("The request is malformed: size limit.",
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'FAKE_CLUSTER')
        mock_check.assert_called_once_with(x_cluster, 2)

    @mock.patch.object(nm.Node, 'load_all')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_collect(self, mock_find, mock_load):
        x_cluster = mock.Mock(id='FAKE_CLUSTER')
        mock_find.return_value = x_cluster
        x_node_1 = mock.Mock(id='NODE1', physical_id='PHYID1')
        x_node_1.to_dict.return_value = {'name': 'node1'}
        x_node_1.get_details.return_value = {'ip': '1.2.3.4'}
        x_node_2 = mock.Mock(id='NODE2', physical_id='PHYID2')
        x_node_2.to_dict.return_value = {'name': 'node2'}
        x_node_2.get_details.return_value = {'ip': '5.6.7.8'}
        mock_load.return_value = [x_node_1, x_node_2]

        res = self.eng.cluster_collect(self.ctx, 'CLUSTER_ID', 'details.ip')

        self.assertIn('cluster_attributes', res)
        self.assertIn({'id': 'NODE1', 'value': '1.2.3.4'},
                      res['cluster_attributes'])
        self.assertIn({'id': 'NODE2', 'value': '5.6.7.8'},
                      res['cluster_attributes'])
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER_ID')
        mock_load.assert_called_once_with(self.ctx, cluster_id='FAKE_CLUSTER',
                                          project_safe=True)
        x_node_1.to_dict.assert_called_once_with()
        x_node_1.get_details.assert_called_once_with(self.ctx)
        x_node_2.to_dict.assert_called_once_with()
        x_node_2.get_details.assert_called_once_with(self.ctx)

    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(common_utils, 'get_path_parser')
    def test_cluster_collect_bad_path(self, mock_parser, mock_find):
        mock_parser.side_effect = exc.BadRequest(msg='Boom')

        err = self.assertRaises(rpc.ExpectedException,
                                self.eng.cluster_collect,
                                self.ctx, 'CLUSTER-ID', 'foo.bar')

        self.assertEqual(exc.BadRequest, err.exc_info[0])
        mock_parser.assert_called_once_with('foo.bar')
        self.assertEqual(0, mock_find.call_count)

    @mock.patch.object(nm.Node, 'load_all')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_collect_cluster_not_found(self, mock_find, mock_load):
        cid = 'FAKE_CLUSTER'
        mock_find.side_effect = exc.ResourceNotFound(type='cluster', id=cid)

        err = self.assertRaises(rpc.ExpectedException,
                                self.eng.cluster_collect,
                                self.ctx, cid, 'foo.bar')
        self.assertEqual(exc.ResourceNotFound, err.exc_info[0])
        mock_find.assert_called_once_with(self.ctx, cid)
        self.assertEqual(0, mock_load.call_count)

    @mock.patch.object(nm.Node, 'load_all')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_collect_no_nodes(self, mock_find, mock_load):
        x_cluster = mock.Mock(id='FAKE_CLUSTER')
        mock_find.return_value = x_cluster
        mock_load.return_value = []

        res = self.eng.cluster_collect(self.ctx, 'CLUSTER_ID', 'bogus')

        self.assertEqual({'cluster_attributes': []}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER_ID')
        mock_load.assert_called_once_with(self.ctx, cluster_id='FAKE_CLUSTER',
                                          project_safe=True)

    @mock.patch.object(nm.Node, 'load_all')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_collect_no_details(self, mock_find, mock_load):
        x_cluster = mock.Mock(id='FAKE_CLUSTER')
        mock_find.return_value = x_cluster
        x_node_1 = mock.Mock(id='NODE1', physical_id=None)
        x_node_1.to_dict.return_value = {'name': 'node1'}
        x_node_2 = mock.Mock(id='NODE2', physical_id=None)
        x_node_2.to_dict.return_value = {'name': 'node2'}
        mock_load.return_value = [x_node_1, x_node_2]

        res = self.eng.cluster_collect(self.ctx, 'CLUSTER_ID', 'name')

        self.assertIn('cluster_attributes', res)
        self.assertIn({'id': 'NODE1', 'value': 'node1'},
                      res['cluster_attributes'])
        self.assertIn({'id': 'NODE2', 'value': 'node2'},
                      res['cluster_attributes'])
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER_ID')
        mock_load.assert_called_once_with(self.ctx, cluster_id='FAKE_CLUSTER',
                                          project_safe=True)
        x_node_1.to_dict.assert_called_once_with()
        self.assertEqual(0, x_node_1.get_details.call_count)
        x_node_2.to_dict.assert_called_once_with()
        self.assertEqual(0, x_node_2.get_details.call_count)

    @mock.patch.object(nm.Node, 'load_all')
    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_collect_no_match(self, mock_find, mock_load):
        x_cluster = mock.Mock(id='FAKE_CLUSTER')
        mock_find.return_value = x_cluster
        x_node_1 = mock.Mock(physical_id=None)
        x_node_1.to_dict.return_value = {'name': 'node1'}
        x_node_2 = mock.Mock(physical_id=None)
        x_node_2.to_dict.return_value = {'name': 'node2'}
        mock_load.return_value = [x_node_1, x_node_2]

        res = self.eng.cluster_collect(self.ctx, 'CLUSTER_ID', 'bogus')

        self.assertEqual({'cluster_attributes': []}, res)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER_ID')
        mock_load.assert_called_once_with(self.ctx, cluster_id='FAKE_CLUSTER',
                                          project_safe=True)
        x_node_1.to_dict.assert_called_once_with()
        self.assertEqual(0, x_node_1.get_details.call_count)
        x_node_2.to_dict.assert_called_once_with()
        self.assertEqual(0, x_node_2.get_details.call_count)

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_check(self, notify, mock_find, mock_action):
        x_cluster = mock.Mock(id='CID')
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'

        params = {'foo': 'bar'}
        result = self.eng.cluster_check(self.ctx, 'CLUSTER', params)

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_CHECK,
            name='cluster_check_CID',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'foo': 'bar'},
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_check_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_check,
                               self.ctx, 'Bogus')

        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual('The cluster (Bogus) could not be found.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')

    @mock.patch.object(am.Action, 'create')
    @mock.patch.object(service.EngineService, 'cluster_find')
    @mock.patch.object(dispatcher, 'start_action')
    def test_cluster_recover(self, notify, mock_find, mock_action):
        x_cluster = mock.Mock(id='CID')
        mock_find.return_value = x_cluster
        mock_action.return_value = 'ACTION_ID'

        params = {'foo': 'bar'}
        result = self.eng.cluster_recover(self.ctx, 'CLUSTER', params)

        self.assertEqual({'action': 'ACTION_ID'}, result)
        mock_find.assert_called_once_with(self.ctx, 'CLUSTER')
        mock_action.assert_called_once_with(
            self.ctx, 'CID', consts.CLUSTER_RECOVER,
            name='cluster_recover_CID',
            cause=am.CAUSE_RPC,
            status=am.Action.READY,
            inputs={'foo': 'bar'},
        )
        notify.assert_called_once_with()

    @mock.patch.object(service.EngineService, 'cluster_find')
    def test_cluster_recover_cluster_not_found(self, mock_find):
        mock_find.side_effect = exc.ResourceNotFound(type='cluster',
                                                     id='Bogus')

        ex = self.assertRaises(rpc.ExpectedException,
                               self.eng.cluster_recover,
                               self.ctx, 'Bogus')
        self.assertEqual(exc.ResourceNotFound, ex.exc_info[0])
        self.assertEqual('The cluster (Bogus) could not be found.',
                         six.text_type(ex.exc_info[1]))
        mock_find.assert_called_once_with(self.ctx, 'Bogus')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.drivers.container import docker_v1 as docker_driver
from senlin.tests.unit.common import base


class TestDocker(base.SenlinTestCase):

    def setUp(self):
        super(TestDocker, self).setUp()

        self._dockerclient = mock.Mock()

    @mock.patch('docker.Client')
    def test_init(self, mock_client):
        mock_client.return_value = self._dockerclient
        url = mock.Mock()
        dockerclient = docker_driver.DockerClient(url)._dockerclient
        self.assertEqual(self._dockerclient, dockerclient)

    @mock.patch('docker.Client')
    def test_container_create(self, mock_client):
        mock_client.return_value = self._dockerclient
        url = mock.Mock()
        dockerclient = docker_driver.DockerClient(url)
        image = mock.Mock()
        dockerclient.container_create(image)
        self._dockerclient.create_container.assert_called_once_with(
            name=None, image=image, command=None)

    @mock.patch('docker.Client')
    def test_container_delete(self, mock_client):
        mock_client.return_value = self._dockerclient
        url = mock.Mock()
        dockerclient = docker_driver.DockerClient(url)
        container = mock.Mock()
        res = dockerclient.container_delete(container)
        self.assertTrue(res)
        self._dockerclient.remove_container.assert_called_once_with(container)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
Client side of the senlin engine RPC API.
'''

from oslo_config import cfg

from senlin.common import consts
from senlin.common import messaging


class EngineClient(object):
    """Client side of the senlin engine rpc API.

    Version History:

      1.0 - Initial version (Mitaka 1.0 release)
      1.1 - Add cluster-collect call.
    """

    BASE_RPC_API_VERSION = '1.0'

    def __init__(self):
        self._client = messaging.get_rpc_client(
            topic=consts.ENGINE_TOPIC,
            server=cfg.CONF.host,
            version=self.BASE_RPC_API_VERSION)

    @staticmethod
    def make_msg(method, **kwargs):
        return method, kwargs

    def call(self, ctxt, msg, version=None):
        method, kwargs = msg
        if version is not None:
            client = self._client.prepare(version=version)
        else:
            client = self._client
        return client.call(ctxt, method, **kwargs)

    def cast(self, ctxt, msg, version=None):
        method, kwargs = msg
        if version is not None:
            client = self._client.prepare(version=version)
        else:
            client = self._client
        return client.cast(ctxt, method, **kwargs)

    def credential_create(self, ctxt, cred, attrs=None):
        return self.call(ctxt, self.make_msg('credential_create', cred=cred,
                                             attrs=attrs))

    def credential_get(self, ctxt, query=None):
        return self.call(ctxt, self.make_msg('credential_get', query=query))

    def credential_update(self, ctxt, cred, attrs=None):
        return self.call(ctxt, self.make_msg('credential_update', cred=cred,
                                             attrs=attrs))

    def profile_type_list(self, ctxt):
        return self.call(ctxt, self.make_msg('profile_type_list'))

    def profile_type_get(self, ctxt, type_name):
        return self.call(ctxt, self.make_msg('profile_type_get',
                                             type_name=type_name))

    def profile_list(self, ctxt, limit=None, marker=None, sort=None,
                     filters=None, project_safe=True):
        return self.call(ctxt,
                         self.make_msg('profile_list', filters=filters,
                                       limit=limit, marker=marker, sort=sort,
                                       project_safe=project_safe))

    def profile_create(self, ctxt, name, spec, metadata):
        return self.call(ctxt,
                         self.make_msg('profile_create', name=name,
                                       spec=spec, metadata=metadata))

    def profile_get(self, ctxt, identity):
        return self.call(ctxt,
                         self.make_msg('profile_get', identity=identity))

    def profile_update(self, ctxt, profile_id, name, metadata):
        return self.call(ctxt,
                         self.make_msg('profile_update',
                                       profile_id=profile_id,
                                       name=name, metadata=metadata))

    def profile_delete(self, ctxt, identity, cast=True):
        rpc_method = self.cast if cast else self.call
        return rpc_method(ctxt,
                          self.make_msg('profile_delete',
                                        identity=identity))

    def profile_validate(self, ctxt, spec):
        return self.call(ctxt,
                         self.make_msg('profile_validate', spec=spec))

    def policy_type_list(self, ctxt):
        return self.call(ctxt, self.make_msg('policy_type_list'))

    def policy_type_get(self, ctxt, type_name):
        return self.call(ctxt, self.make_msg('policy_type_get',
                                             type_name=type_name))

    def policy_list(self, ctxt, limit=None, marker=None, sort=None,
                    filters=None, project_safe=True):
        return self.call(ctxt,
                         self.make_msg('policy_list', filters=filters,
                                       limit=limit, marker=marker, sort=sort,
                                       project_safe=project_safe))

    def policy_create(self, ctxt, name, spec):
        return self.call(ctxt,
                         self.make_msg('policy_create', name=name, spec=spec))

    def policy_get(self, ctxt, identity):
        return self.call(ctxt,
                         self.make_msg('policy_get', identity=identity))

    def policy_update(self, ctxt, identity, name):
        return self.call(ctxt,
                         self.make_msg('policy_update', identity=identity,
                                       name=name))

    def policy_delete(self, ctxt, identity, cast=True):
        rpc_method = self.cast if cast else self.call
        return rpc_method(ctxt,
                          self.make_msg('policy_delete',
                                        identity=identity))

    def policy_validate(self, ctxt, spec):
        return self.call(ctxt,
                         self.make_msg('policy_validate', spec=spec))

    def cluster_list(self, ctxt, limit=None, marker=None, sort=None,
                     filters=None, project_safe=True):
        return self.call(ctxt,
                         self.make_msg('cluster_list', filters=filters,
                                       limit=limit, marker=marker, sort=sort,
                                       project_safe=project_safe))

    def cluster_get(self, ctxt, identity):
        return self.call(ctxt,
                         self.make_msg('cluster_get', identity=identity))

    def cluster_create(self, ctxt, name, desired_capacity, profile_id,
                       min_size=None, max_size=None, metadata=None,
                       timeout=None):
        return self.call(ctxt, self.make_msg('cluster_create',
                                             name=name,
                                             desired_capacity=desired_capacity,
                                             profile_id=profile_id,
                                             min_size=min_size,
                                             max_size=max_size,
                                             metadata=metadata,
                                             timeout=timeout))

    def cluster_add_nodes(self, ctxt, identity, nodes):
        return self.call(ctxt, self.make_msg('cluster_add_nodes',
                                             identity=identity,
                                             nodes=nodes))

    def cluster_del_nodes(self, ctxt, identity, nodes):
        return self.call(ctxt, self.make_msg('cluster_del_nodes',
                                             identity=identity,
                                             nodes=nodes))

    def cluster_resize(self, ctxt, identity, adj_type=None, number=None,
                       min_size=None, max_size=None, min_step=None,
                       strict=True):
        return self.call(ctxt, self.make_msg('cluster_resize',
                                             identity=identity,
                                             adj_type=adj_type,
                                             number=number,
                                             min_size=min_size,
                                             max_size=max_size,
                                             min_step=min_step,
                                             strict=strict))

    def cluster_scale_out(self, ctxt, identity, count=None):
        return self.call(ctxt, self.make_msg('cluster_scale_out',
                                             identity=identity,
                                             count=count))

    def cluster_scale_in(self, ctxt, identity, count=None):
        return self.call(ctxt, self.make_msg('cluster_scale_in',
                                             identity=identity,
                                             count=count))

    def cluster_update(self, ctxt, identity, name=None, profile_id=None,
                       metadata=None, timeout=None):
        return self.call(ctxt, self.make_msg('cluster_update',
                                             identity=identity, name=name,
                                             profile_id=profile_id,
                                             metadata=metadata,
                                             timeout=timeout))

    def cluster_delete(self, ctxt, identity, cast=True):
        rpc_method = self.cast if cast else self.call
        return rpc_method(ctxt,
                          self.make_msg('cluster_delete',
                                        identity=identity))

    def cluster_collect(self, ctxt, identity, path, project_safe=True):
        return self.call(ctxt, self.make_msg('cluster_collect',
                                             identity=identity, path=path,
                                             project_safe=project_safe),
                         version='1.1')

    def cluster_check(self, ctxt, identity, params=None):
        return self.call(ctxt, self.make_msg('cluster_check',
                                             identity=identity,
                                             params=params))

    def cluster_recover(self, ctxt, identity, params=None):
        return self.call(ctxt, self.make_msg('cluster_recover',
                                             identity=identity,
                                             params=params))

    def node_list(self, ctxt, cluster_id=None, limit=None, marker=None,
                  sort=None, filters=None, project_safe=True):
        return self.call(ctxt,
                         self.make_msg('node_list', cluster_id=cluster_id,
                                       limit=limit, marker=marker, sort=sort,
                                       filters=filters,
                                       project_safe=project_safe))

    def node_create(self, ctxt, name, cluster_id, profile_id, role, metadata):
        return self.call(ctxt,
                         self.make_msg('node_create', name=name,
                                       profile_id=profile_id,
                                       cluster_id=cluster_id,
                                       role=role, metadata=metadata))

    def node_get(self, ctxt, identity, show_details=False):
        return self.call(ctxt,
                         self.make_msg('node_get', identity=identity,
                                       show_details=show_details))

    def node_update(self, ctxt, identity, name, profile_id, role, metadata):
        return self.call(ctxt,
                         self.make_msg('node_update', identity=identity,
                                       name=name, profile_id=profile_id,
                                       role=role, metadata=metadata))

    def node_delete(self, ctxt, identity, cast=True):
        rpc_method = self.cast if cast else self.call
        return rpc_method(ctxt,
                          self.make_msg('node_delete', identity=identity))

    def node_check(self, ctxt, identity, params=None):
        return self.call(ctxt, self.make_msg('node_check',
                                             identity=identity,
                                             params=params))

    def node_recover(self, ctxt, identity, params=None):
        return self.call(ctxt, self.make_msg('node_recover',
                                             identity=identity,
                                             params=params))

    def action_list(self, ctxt, filters=None, limit=None, marker=None,
                    sort=None, project_safe=True):
        return self.call(ctxt,
                         self.make_msg('action_list', filters=filters,
                                       limit=limit, marker=marker,
                                       sort=sort, project_safe=project_safe))

    def cluster_policy_list(self, ctxt, cluster_id, filters=None, sort=None):
        return self.call(ctxt, self.make_msg('cluster_policy_list',
                                             identity=cluster_id,
                                             filters=filters, sort=sort))

    def cluster_policy_attach(self, ctxt, cluster_id, policy_id, enabled=True):
        return self.call(ctxt, self.make_msg('cluster_policy_attach',
                                             identity=cluster_id,
                                             policy=policy_id,
                                             enabled=enabled))

    def cluster_policy_detach(self, ctxt, cluster_id, policy_id):
        return self.call(ctxt, self.make_msg('cluster_policy_detach',
                                             identity=cluster_id,
                                             policy=policy_id))

    def cluster_policy_get(self, ctxt, cluster_id, policy_id):
        return self.call(ctxt, self.make_msg('cluster_policy_get',
                                             identity=cluster_id,
                                             policy_id=policy_id))

    def cluster_policy_update(self, ctxt, cluster_id, policy_id, enabled=None):
        return self.call(ctxt, self.make_msg('cluster_policy_update',
                                             identity=cluster_id,
                                             policy=policy_id,
                                             enabled=enabled))

    def action_create(self, ctxt, name, cluster, action, params):
        return self.call(ctxt,
                         self.make_msg('action_create',
                                       name=name, cluster=cluster,
                                       action=action, params=params))

    def action_get(self, ctxt, identity):
        return self.call(ctxt,
                         self.make_msg('action_get', identity=identity))

    def receiver_list(self, ctxt, limit=None, marker=None, filters=None,
                      sort=None, project_safe=True):
        return self.call(ctxt,
                         self.make_msg('receiver_list', filters=filters,
                                       limit=limit, marker=marker,
                                       sort=sort, project_safe=project_safe))

    def receiver_create(self, ctxt, name, type_name, cluster_id, action,
                        actor=None, params=None):
        return self.call(ctxt,
                         self.make_msg('receiver_create',
                                       name=name, type_name=type_name,
                                       cluster_id=cluster_id, action=action,
                                       actor=actor, params=params))

    def receiver_get(self, ctxt, identity, project_safe=True):
        return self.call(ctxt, self.make_msg('receiver_get', identity=identity,
                                             project_safe=project_safe))

    def receiver_delete(self, ctxt, identity, cast=True):
        rpc_method = self.cast if cast else self.call
        return rpc_method(ctxt, self.make_msg('receiver_delete',
                                              identity=identity))

    def webhook_trigger(self, ctxt, identity, params=None):
        return self.call(ctxt,
                         self.make_msg('webhook_trigger', identity=identity,
                                       params=params))

    def event_list(self, ctxt, filters=None, limit=None, marker=None,
                   sort=None, project_safe=True):
        return self.call(ctxt,
                         self.make_msg('event_list', filters=filters,
                                       limit=limit, marker=marker,
                                       sort=sort, project_safe=project_safe))

    def event_get(self, ctxt, identity):
        return self.call(ctxt,
                         self.make_msg('event_get', identity=identity))

    def get_revision(self, ctxt):
        return self.call(ctxt, self.make_msg('get_revision'))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Cluster object."""

from senlin.db import api as db_api
from senlin.objects import base
from senlin.objects import fields


@base.SenlinObjectRegistry.register
class Cluster(base.SenlinObject, base.VersionedObjectDictCompat):
    """Senlin cluster object."""

    fields = {
        'id': fields.UUIDField(),
        'name': fields.StringField(),
        'profile_id': fields.UUIDField(),
        'parent': fields.UUIDField(nullable=True),
        'init_at': fields.DateTimeField(),
        'created_at': fields.DateTimeField(nullable=True),
        'updated_at': fields.DateTimeField(nullable=True),
        'min_size': fields.IntegerField(nullable=True),
        'max_size': fields.IntegerField(nullable=True),
        'desired_capacity': fields.IntegerField(nullable=True),
        'next_index': fields.IntegerField(nullable=True),
        'timeout': fields.IntegerField(nullable=True),
        'status': fields.StringField(),
        'status_reason': fields.StringField(nullable=True),
        'metadata': fields.JsonField(nullable=True),
        'data': fields.JsonField(nullable=True),
        'user': fields.StringField(),
        'project': fields.StringField(),
        'domain': fields.StringField(nullable=True),
        'dependents': fields.JsonField(nullable=True),
    }

    @classmethod
    def create(cls, context, values):
        values = cls._transpose_metadata(values)
        obj = db_api.cluster_create(context, values)
        return cls._from_db_object(context, cls(context), obj)

    @classmethod
    def get(cls, context, cluster_id, **kwargs):
        obj = db_api.cluster_get(context, cluster_id, **kwargs)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def get_by_name(cls, context, name, **kwargs):
        obj = db_api.cluster_get_by_name(context, name, **kwargs)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def get_by_short_id(cls, context, short_id, **kwargs):
        obj = db_api.cluster_get_by_short_id(context, short_id, **kwargs)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def get_all(cls, context, **kwargs):
        objs = db_api.cluster_get_all(context, **kwargs)
        return [cls._from_db_object(context, cls(), obj) for obj in objs]

    @classmethod
    def get_next_index(cls, context, cluster_id):
        return db_api.cluster_next_index(context, cluster_id)

    @classmethod
    def count_all(cls, context, **kwargs):
        return db_api.cluster_count_all(context, **kwargs)

    @classmethod
    def update(cls, context, obj_id, values):
        values = cls._transpose_metadata(values)
        return db_api.cluster_update(context, obj_id, values)

    @classmethod
    def delete(cls, context, obj_id):
        db_api.cluster_delete(context, obj_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

RPC_ATTRS = (
    ENGINE_TOPIC,
    ENGINE_DISPATCHER_TOPIC,
    ENGINE_HEALTH_MGR_TOPIC,
    RPC_API_VERSION,
) = (
    'senlin-engine',
    'engine-dispatcher',
    'engine-health-mgr',
    '1.1',
)

RPC_PARAMS = (
    PARAM_LIMIT, PARAM_MARKER, PARAM_GLOBAL_PROJECT,
    PARAM_SHOW_DETAILS, PARAM_SORT,
) = (
    'limit', 'marker', 'global_project',
    'show_details', 'sort',
)

ACTION_NAMES = (
    CLUSTER_CREATE, CLUSTER_DELETE, CLUSTER_UPDATE,
    CLUSTER_ADD_NODES, CLUSTER_DEL_NODES, CLUSTER_RESIZE,
    CLUSTER_CHECK, CLUSTER_RECOVER,
    CLUSTER_SCALE_OUT, CLUSTER_SCALE_IN,
    CLUSTER_ATTACH_POLICY, CLUSTER_DETACH_POLICY, CLUSTER_UPDATE_POLICY,

    NODE_CREATE, NODE_DELETE, NODE_UPDATE,
    NODE_JOIN, NODE_LEAVE,
    NODE_CHECK, NODE_RECOVER
) = (
    'CLUSTER_CREATE', 'CLUSTER_DELETE', 'CLUSTER_UPDATE',
    'CLUSTER_ADD_NODES', 'CLUSTER_DEL_NODES', 'CLUSTER_RESIZE',
    'CLUSTER_CHECK', 'CLUSTER_RECOVER',
    'CLUSTER_SCALE_OUT', 'CLUSTER_SCALE_IN',
    'CLUSTER_ATTACH_POLICY', 'CLUSTER_DETACH_POLICY', 'CLUSTER_UPDATE_POLICY',

    'NODE_CREATE', 'NODE_DELETE', 'NODE_UPDATE',
    'NODE_JOIN', 'NODE_LEAVE',
    'NODE_CHECK', 'NODE_RECOVER'
)

ADJUSTMENT_PARAMS = (
    ADJUSTMENT_TYPE, ADJUSTMENT_NUMBER, ADJUSTMENT_MIN_STEP,
    ADJUSTMENT_MIN_SIZE, ADJUSTMENT_MAX_SIZE, ADJUSTMENT_STRICT,
) = (
    'adjustment_type', 'number', 'min_step',
    'min_size', 'max_size', 'strict',
)

ADJUSTMENT_TYPES = (
    EXACT_CAPACITY, CHANGE_IN_CAPACITY, CHANGE_IN_PERCENTAGE,
) = (
    'EXACT_CAPACITY', 'CHANGE_IN_CAPACITY', 'CHANGE_IN_PERCENTAGE',
)

CLUSTER_ATTRS = (
    CLUSTER_NAME, CLUSTER_PROFILE, CLUSTER_DESIRED_CAPACITY,
    CLUSTER_MIN_SIZE, CLUSTER_MAX_SIZE, CLUSTER_ID,
    CLUSTER_DOMAIN, CLUSTER_PROJECT, CLUSTER_USER,
    CLUSTER_INIT_AT, CLUSTER_CREATED_AT, CLUSTER_UPDATED_AT,
    CLUSTER_STATUS, CLUSTER_STATUS_REASON, CLUSTER_TIMEOUT,
    CLUSTER_METADATA,
) = (
    'name', 'profile_id', 'desired_capacity',
    'min_size', 'max_size', 'id',
    'domain', 'project', 'user',
    'init_at', 'created_at', 'updated_at',
    'status', 'status_reason', 'timeout',
    'metadata',
)

CLUSTER_SORT_KEYS = [
    CLUSTER_NAME, CLUSTER_STATUS,
    CLUSTER_INIT_AT, CLUSTER_CREATED_AT, CLUSTER_UPDATED_AT,
]

DETECTION_TYPES = (
    VM_LIFECYCLE_EVENTS, NODE_STATUS_POLLING, LB_STATUS_POLLING,
) = (
    'VM_LIFECYCLE_EVENTS', 'NODE_STATUS_POLLING', 'LB_STATUS_POLLING',
)

NODE_ATTRS = (
    NODE_INDEX, NODE_NAME, NODE_PROFILE_ID, NODE_CLUSTER_ID,
    NODE_INIT_AT, NODE_CREATED_AT, NODE_UPDATED_AT,
    NODE_STATUS, NODE_ROLE, NODE_METADATA,
) = (
    'index', 'name', 'profile_id', 'cluster_id',
    'init_at', 'created_at', 'updated_at',
    'status', 'role', 'metadata',
)

NODE_SORT_KEYS = [
    NODE_INDEX, NODE_NAME, NODE_STATUS,
    NODE_INIT_AT, NODE_CREATED_AT, NODE_UPDATED_AT,
]

PROFILE_ATTRS = (
    PROFILE_ID, PROFILE_NAME, PROFILE_TYPE,
    PROFILE_CREATED_AT, PROFILE_UPDATED_AT,
    PROFILE_SPEC, PROFILE_METADATA, PROFILE_CONTEXT,
) = (
    'id', 'name', 'type',
    'created_at', 'updated_at',
    'spec', 'metadata', 'context',
)

PROFILE_SORT_KEYS = [
    PROFILE_TYPE, PROFILE_NAME, PROFILE_CREATED_AT, PROFILE_UPDATED_AT,
]

POLICY_ATTRS = (
    POLICY_ID, POLICY_NAME, POLICY_TYPE, POLICY_SPEC,
    POLICY_CREATED_AT, POLICY_UPDATED_AT,
) = (
    'id', 'name', 'type', 'spec',
    'created_at', 'updated_at',
)

POLICY_SORT_KEYS = [
    POLICY_TYPE, POLICY_NAME,
    POLICY_CREATED_AT, POLICY_UPDATED_AT,
]

CLUSTER_POLICY_ATTRS = (
    CP_POLICY_ID, CP_ENABLED, CP_PRIORITY,
) = (
    'policy_id', 'enabled', 'priority'
)

CLUSTER_POLICY_SORT_KEYS = [
    CP_ENABLED, CP_PRIORITY,
]

EVENT_ATTRS = (
    EVENT_TIMESTAMP, EVENT_OBJ_ID, EVENT_OBJ_NAME, EVENT_OBJ_TYPE,
    EVENT_USER, EVENT_ACTION, EVENT_STATUS, EVENT_STATUS_REASON,
    EVENT_LEVEL, EVENT_CLUSTER_ID,
) = (
    'timestamp', 'oid', 'oname', 'otype',
    'user', 'action', 'status', 'status_reason',
    'level', 'cluster_id',
)

EVENT_SORT_KEYS = [
    EVENT_TIMESTAMP, EVENT_LEVEL, EVENT_OBJ_TYPE, EVENT_OBJ_NAME,
    EVENT_USER, EVENT_ACTION, EVENT_STATUS,
]

ACTION_ATTRS = (
    ACTION_NAME, ACTION_TARGET, ACTION_ACTION, ACTION_CAUSE,
    ACTION_INTERVAL, ACTION_START_TIME, ACTION_END_TIME,
    ACTION_TIMEOUT, ACTION_STATUS, ACTION_STATUS_REASON,
    ACTION_INPUTS, ACTION_OUTPUTS, ACTION_DEPENDS_ON, ACTION_DEPENDED_BY,
    ACTION_CREATED_AT, ACTION_UPDATED_AT,
) = (
    'name', 'target', 'action', 'cause',
    'interval', 'start_time', 'end_time',
    'timeout', 'status', 'status_reason',
    'inputs', 'outputs', 'depends_on', 'depended_by',
    'created_at', 'updated_at',
)

ACTION_SORT_KEYS = [
    ACTION_NAME, ACTION_TARGET, ACTION_ACTION, ACTION_CREATED_AT,
    ACTION_STATUS,
]

RECEIVER_TYPES = (
    RECEIVER_WEBHOOK, RECEIVER_MESSAGE,
) = (
    'webhook', 'message',
)

RECEIVER_ATTRS = (
    RECEIVER_NAME, RECEIVER_TYPE, RECEIVER_CLUSTER, RECEIVER_CLUSTER_ID,
    RECEIVER_CREATED_AT, RECEIVER_UPDATED_AT,
    RECEIVER_ACTOR, RECEIVER_ACTION, RECEIVER_PARAMS, RECEIVER_CHANNEL,
) = (
    'name', 'type', 'cluster', 'cluster_id',
    'created_at', 'updated_at',
    'actor', 'action', 'params', 'channel',
)

RECEIVER_SORT_KEYS = [
    RECEIVER_NAME, RECEIVER_TYPE, RECEIVER_ACTION, RECEIVER_CLUSTER_ID,
    RECEIVER_CREATED_AT,
]

CLUSTER_DEFAULT_VALUE = (
    CLUSTER_DEFAULT_MIN_SIZE, CLUSTER_DEFAULT_MAX_SIZE
) = (
    0, -1
)

# Note: This is a copy of action status definition defined in
# senlin.engine.actions.base module.
ACTION_STATUSES = (
    ACTION_INIT, ACTION_WAITING, ACTION_READY, ACTION_RUNNING,
    ACTION_SUCCEEDED, ACTION_FAILED, ACTION_CANCELLED
) = (
    'INIT', 'WAITING', 'READY', 'RUNNING',
    'SUCCEEDED', 'FAILED', 'CANCELLED',
)

EVENT_LEVELS = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
}

RECOVER_OPERATIONS = (
    RECOVER_RECREATE, RECOVER_REBUILD,
) = (
    'RECREATE', 'REBUILD',
)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Webhook endpoint for Senlin v1 ReST API.
"""

from webob import exc

from senlin.api.common import util
from senlin.api.common import wsgi
from senlin.common import consts
from senlin.common.i18n import _
from senlin.common import utils


class ReceiverData(object):
    """The data accompanying a POST request to create a receiver."""

    PARAMS = (consts.RECEIVER_NAME, consts.RECEIVER_TYPE,
              consts.RECEIVER_CLUSTER, consts.RECEIVER_ACTION,
              consts.RECEIVER_ACTOR, consts.RECEIVER_PARAMS)

    def __init__(self, data):
        self.data = data

    def name(self):
        if consts.RECEIVER_NAME not in self.data:
            raise exc.HTTPBadRequest(_("Missing 'name' in request."))
        return self.data.get(consts.RECEIVER_NAME, None)

    def type_name(self):
        if consts.RECEIVER_TYPE not in self.data:
            raise exc.HTTPBadRequest(_("Missing 'type' in request."))
        return self.data[consts.RECEIVER_TYPE]

    def cluster_id(self):
        if self.data.get(consts.RECEIVER_TYPE) == consts.RECEIVER_WEBHOOK:
            if consts.RECEIVER_CLUSTER_ID not in self.data:
                raise exc.HTTPBadRequest(_("Missing 'cluster_id' in request."))
        return self.data.get(consts.RECEIVER_CLUSTER_ID, None)

    def action(self):
        if self.data.get(consts.RECEIVER_TYPE) == consts.RECEIVER_WEBHOOK:
            if consts.RECEIVER_ACTION not in self.data:
                raise exc.HTTPBadRequest(_("Missing 'action' in request."))
        return self.data.get(consts.RECEIVER_ACTION, None)

    def actor(self):
        return self.data.get(consts.RECEIVER_ACTOR, None)

    def params(self):
        return self.data.get(consts.RECEIVER_PARAMS, None)


class ReceiverController(wsgi.Controller):
    """WSGI controller for receiver resource in Senlin v1 API."""

    REQUEST_SCOPE = 'receivers'

    @util.policy_enforce
    def index(self, req):
        filter_whitelist = {
            consts.RECEIVER_NAME: 'mixed',
            consts.RECEIVER_TYPE: 'mixed',
            consts.RECEIVER_CLUSTER_ID: 'mixed',
            consts.RECEIVER_ACTION: 'mixed',
        }
        param_whitelist = {
            consts.PARAM_LIMIT: 'single',
            consts.PARAM_MARKER: 'single',
            consts.PARAM_SORT: 'single',
            consts.PARAM_GLOBAL_PROJECT: 'single',
        }
        for key in req.params.keys():
            if (key not in param_whitelist.keys() and key not in
                    filter_whitelist.keys()):
                raise exc.HTTPBadRequest(_('Invalid parameter %s') % key)
        params = util.get_allowed_params(req.params, param_whitelist)
        filters = util.get_allowed_params(req.params, filter_whitelist)

        key = consts.PARAM_LIMIT
        if key in params:
            params[key] = utils.parse_int_param(key, params[key])

        key = consts.PARAM_GLOBAL_PROJECT
        if key in params:
            show_global = utils.parse_bool_param(key, params[key])
            del params[key]
            params['project_safe'] = not show_global

        if not filters:
            filters = None

        receivers = self.rpc_client.receiver_list(req.context, filters=filters,
                                                  **params)

        return {'receivers': receivers}

    @util.policy_enforce
    def create(self, req, body):
        data = body.get('receiver')
        if data is None:
            raise exc.HTTPBadRequest(_("Malformed request data, missing "
                                       "'receiver' key in request body."))

        data = ReceiverData(data)

        result = self.rpc_client.receiver_create(req.context,
                                                 data.name(),
                                                 data.type_name(),
                                                 data.cluster_id(),
                                                 data.action(),
                                                 data.actor(),
                                                 data.params())
        return {'receiver': result}

    @util.policy_enforce
    def get(self, req, receiver_id):
        receiver = self.rpc_client.receiver_get(req.context, receiver_id)
        return {'receiver': receiver}

    @util.policy_enforce
    def delete(self, req, receiver_id):
        self.rpc_client.receiver_delete(req.context, receiver_id, cast=False)
        raise exc.HTTPNoContent()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.common import consts
from senlin.common import scaleutils as su
from senlin.engine import cluster as cm
from senlin.objects import cluster as co
from senlin.policies import deletion_policy as dp
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestDeletionPolicy(base.SenlinTestCase):

    def setUp(self):
        super(TestDeletionPolicy, self).setUp()
        self.context = utils.dummy_context()
        self.spec = {
            'type': 'senlin.policy.deletion',
            'version': '1.0',
            'properties': {
                'criteria': 'OLDEST_FIRST',
                'destroy_after_deletion': True,
                'grace_period': 60,
                'reduce_desired_capacity': False
            }
        }

    def test_policy_init(self):
        policy = dp.DeletionPolicy('test-policy', self.spec)

        self.assertIsNone(policy.id)
        self.assertEqual('test-policy', policy.name)
        self.assertEqual('senlin.policy.deletion-1.0', policy.type)
        self.assertEqual('OLDEST_FIRST', policy.criteria)
        self.assertTrue(policy.destroy_after_deletion)
        self.assertEqual(60, policy.grace_period)
        self.assertFalse(policy.reduce_desired_capacity)

    @mock.patch.object(su, 'nodes_by_random')
    def test__victims_by_regions_random(self, mock_select):
        cluster = mock.Mock()
        node1 = mock.Mock(id=1)
        node2 = mock.Mock(id=2)
        node3 = mock.Mock(id=3)
        cluster.nodes_by_region.side_effect = [
            [node1], [node2, node3]
        ]

        mock_select.side_effect = [['1'], ['2', '3']]

        self.spec['properties']['criteria'] = 'RANDOM'
        policy = dp.DeletionPolicy('test-policy', self.spec)

        res = policy._victims_by_regions(cluster, {'R1': 1, 'R2': 2})
        self.assertEqual(['1', '2', '3'], res)
        mock_select.assert_has_calls([
            mock.call([node1], 1),
            mock.call([node2, node3], 2)
        ])
        cluster.nodes_by_region.assert_has_calls([
            mock.call('R1'), mock.call('R2')])

    @mock.patch.object(su, 'nodes_by_profile_age')
    def test__victims_by_regions_profile_age(self, mock_select):
        cluster = mock.Mock()
        node1 = mock.Mock(id=1)
        node2 = mock.Mock(id=2)
        node3 = mock.Mock(id=3)
        cluster.nodes_by_region.side_effect = [
            [node1], [node2, node3]
        ]

        mock_select.side_effect = [['1'], ['2', '3']]

        self.spec['properties']['criteria'] = 'OLDEST_PROFILE_FIRST'
        policy = dp.DeletionPolicy('test-policy', self.spec)

        res = policy._victims_by_regions(cluster, {'R1': 1, 'R2': 2})
        self.assertEqual(['1', '2', '3'], res)
        mock_select.assert_has_calls([
            mock.call([node1], 1),
            mock.call([node2, node3], 2)
        ])
        cluster.nodes_by_region.assert_has_calls([
            mock.call('R1'), mock.call('R2')])

    @mock.patch.object(su, 'nodes_by_age')
    def test__victims_by_regions_age_oldest(self, mock_select):
        cluster = mock.Mock()
        node1 = mock.Mock(id=1)
        node2 = mock.Mock(id=2)
        node3 = mock.Mock(id=3)
        cluster.nodes_by_region.side_effect = [
            [node1], [node2, node3]
        ]

        mock_select.side_effect = [['1'], ['2', '3']]

        self.spec['properties']['criteria'] = 'OLDEST_FIRST'
        policy = dp.DeletionPolicy('test-policy', self.spec)

        res = policy._victims_by_regions(cluster, {'R1': 1, 'R2': 2})
        self.assertEqual(['1', '2', '3'], res)
        mock_select.assert_has_calls([
            mock.call([node1], 1, True),
            mock.call([node2, node3], 2, True)
        ])
        cluster.nodes_by_region.assert_has_calls([
            mock.call('R1'), mock.call('R2')])

    @mock.patch.object(su, 'nodes_by_age')
    def test__victims_by_regions_age_youngest(self, mock_select):
        cluster = mock.Mock()
        node1 = mock.Mock(id=1)
        node2 = mock.Mock(id=2)
        node3 = mock.Mock(id=3)
        cluster.nodes_by_region.side_effect = [
            [node1], [node2, node3]
        ]

        mock_select.side_effect = [['1'], ['2', '3']]

        self.spec['properties']['criteria'] = 'YOUNGEST_FIRST'
        policy = dp.DeletionPolicy('test-policy', self.spec)

        res = policy._victims_by_regions(cluster, {'R1': 1, 'R2': 2})
        self.assertEqual(['1', '2', '3'], res)
        mock_select.assert_has_calls([
            mock.call([node1], 1, False),
            mock.call([node2, node3], 2, False)
        ])
        cluster.nodes_by_region.assert_has_calls([
            mock.call('R1'), mock.call('R2')])

    @mock.patch.object(su, 'nodes_by_random')
    def test__victims_by_zones_random(self, mock_select):
        cluster = mock.Mock()
        node1 = mock.Mock(id=1)
        node2 = mock.Mock(id=2)
        node3 = mock.Mock(id=3)
        cluster.nodes_by_zone.side_effect = [
            [node1], [node2, node3]
        ]

        mock_select.side_effect = [['1'], ['3']]

        self.spec['properties']['criteria'] = 'RANDOM'
        policy = dp.DeletionPolicy('test-policy', self.spec)

        res = policy._victims_by_zones(cluster, {'AZ1': 1, 'AZ2': 1})
        self.assertEqual(['1', '3'], res)
        mock_select.assert_has_calls([
            mock.call([node1], 1),
            mock.call([node2, node3], 1)
        ])
        cluster.nodes_by_zone.assert_has_calls(
            [mock.call('AZ1'), mock.call('AZ2')],
        )

    @mock.patch.object(su, 'nodes_by_profile_age')
    def test__victims_by_zones_profile_age(self, mock_select):
        cluster = mock.Mock()
        node1 = mock.Mock(id=1)
        node2 = mock.Mock(id=2)
        node3 = mock.Mock(id=3)
        cluster.nodes_by_zone.side_effect = [
            [node1], [node2, node3]
        ]

        mock_select.side_effect = [['1'], ['2']]

        self.spec['properties']['criteria'] = 'OLDEST_PROFILE_FIRST'
        policy = dp.DeletionPolicy('test-policy', self.spec)

        res = policy._victims_by_zones(cluster, {'AZ1': 1, 'AZ2': 1})
        self.assertEqual(['1', '2'], res)
        mock_select.assert_has_calls(
            [
                mock.call([node1], 1),
                mock.call([node2, node3], 1)
            ],
        )
        cluster.nodes_by_zone.assert_has_calls(
            [mock.call('AZ1'), mock.call('AZ2')],
        )

    @mock.patch.object(su, 'nodes_by_age')
    def test__victims_by_zones_age_oldest(self, mock_select):
        cluster = mock.Mock()
        node1 = mock.Mock(id=1)
        node2 = mock.Mock(id=2)
        node3 = mock.Mock(id=3)
        cluster.nodes_by_zone.side_effect = [
            [node1], [node2, node3]
        ]

        mock_select.side_effect = [['1'], ['3']]

        self.spec['properties']['criteria'] = 'OLDEST_FIRST'
        policy = dp.DeletionPolicy('test-policy', self.spec)

        res = policy._victims_by_zones(cluster, {'AZ1': 1, 'AZ8': 1})
        self.assertEqual(['1', '3'], res)
        mock_select.assert_has_calls([
            mock.call([node1], 1, True),
            mock.call([node2, node3], 1, True)
        ])
        cluster.nodes_by_zone.assert_has_calls(
            [mock.call('AZ1'), mock.call('AZ8')],
        )

    @mock.patch.object(su, 'nodes_by_age')
    def test__victims_by_zones_age_youngest(self, mock_select):
        cluster = mock.Mock()
        node1 = mock.Mock(id=1)
        node2 = mock.Mock(id=3)
        node3 = mock.Mock(id=5)
        cluster.nodes_by_zone.side_effect = [
            [node1], [node2, node3]
        ]

        mock_select.side_effect = [['1'], ['3', '5']]

        self.spec['properties']['criteria'] = 'YOUNGEST_FIRST'
        policy = dp.DeletionPolicy('test-policy', self.spec)

        res = policy._victims_by_zones(cluster, {'AZ5': 1, 'AZ6': 2})
        self.assertEqual(['1', '3', '5'], res)
        mock_select.assert_has_calls(
            [
                mock.call([node1], 1, False),
                mock.call([node2, node3], 2, False)
            ],
        )
        cluster.nodes_by_zone.assert_has_calls(
            [mock.call('AZ5'), mock.call('AZ6')],
        )

    def test__update_action_clean(self):
        action = mock.Mock()
        action.data = {}

        policy = dp.DeletionPolicy('test-policy', self.spec)

        policy._update_action(action, ['N1', 'N2'])

        pd = {
            'status': 'OK',
            'reason': 'Candidates generated',
            'deletion': {
                'count': 2,
                'candidates': ['N1', 'N2'],
                'destroy_after_deletion': True,
                'grace_period': 60,
                'reduce_desired_capacity': False,
            }
        }
        self.assertEqual(pd, action.data)
        action.store.assert_called_with(action.context)

    def test__update_action_override(self):
        action = mock.Mock()
        action.data = {
            'deletion': {
                'count': 3,
            }
        }

        policy = dp.DeletionPolicy('test-policy', self.spec)

        policy._update_action(action, ['N1', 'N2'])

        pd = {
            'status': 'OK',
            'reason': 'Candidates generated',
            'deletion': {
                'count': 2,
                'candidates': ['N1', 'N2'],
                'destroy_after_deletion': True,
                'grace_period': 60,
                'reduce_desired_capacity': False,
            }
        }
        self.assertEqual(pd, action.data)
        action.store.assert_called_with(action.context)

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    def test_pre_op_del_nodes(self, mock_update):
        action = mock.Mock()
        action.context = self.context
        action.inputs = {
            'count': 2,
            'candidates': ['N1', 'N2'],
        }
        action.data = {}

        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)
        mock_update.assert_called_once_with(action, ['N1', 'N2'])

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    def test_pre_op_node_delete(self, mock_update):
        action = mock.Mock()
        action.action = consts.NODE_DELETE
        action.context = self.context
        action.node.id = 'NODE_ID'
        action.inputs = {}
        action.data = {}

        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)
        mock_update.assert_called_once_with(action, ['NODE_ID'])

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'nodes_by_age')
    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_with_count_decisions(self, mock_load, mock_select,
                                         mock_update):
        action = mock.Mock()
        action.context = self.context
        action.inputs = {}
        action.data = {'deletion': {'count': 2}}

        cluster = mock.Mock(nodes=['a', 'b', 'c'])
        mock_load.return_value = cluster
        mock_select.return_value = ['NODE1', 'NODE2']

        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)

        mock_update.assert_called_once_with(action, ['NODE1', 'NODE2'])
        mock_load.assert_called_once_with(action.context,
                                          dbcluster=None, cluster_id='FAKE_ID')
        mock_select.assert_called_once_with(cluster.nodes, 2, True)

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(dp.DeletionPolicy, '_victims_by_regions')
    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_with_region_decisions(self, mock_load, mock_select,
                                          mock_update):
        action = mock.Mock()
        action.context = self.context
        action.inputs = {}
        action.data = {
            'deletion': {
                'count': 2,
                'regions': {
                    'R1': 1,
                    'R2': 1
                }
            }
        }

        cluster = mock.Mock(nodes=['a', 'b', 'c'])
        mock_load.return_value = cluster
        mock_select.return_value = ['NODE1', 'NODE2']

        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)

        mock_update.assert_called_once_with(action, ['NODE1', 'NODE2'])
        mock_load.assert_called_once_with(action.context,
                                          dbcluster=None, cluster_id='FAKE_ID')
        mock_select.assert_called_once_with(cluster, {'R1': 1, 'R2': 1})

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(dp.DeletionPolicy, '_victims_by_zones')
    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_with_zone_decisions(self, mock_load, mock_select,
                                        mock_update):
        action = mock.Mock()
        action.context = self.context
        action.inputs = {}
        action.data = {
            'deletion': {
                'count': 2,
                'zones': {
                    'AZ1': 1,
                    'AZ2': 1
                }
            }
        }

        cluster = mock.Mock(nodes=['a', 'b', 'c'])
        mock_load.return_value = cluster
        mock_select.return_value = ['NODE1', 'NODE2']

        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)

        mock_update.assert_called_once_with(action, ['NODE1', 'NODE2'])
        mock_load.assert_called_once_with(action.context,
                                          dbcluster=None, cluster_id='FAKE_ID')
        mock_select.assert_called_once_with(cluster, {'AZ1': 1, 'AZ2': 1})

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'nodes_by_age')
    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_scale_in_with_count(self, mock_load, mock_select,
                                        mock_update):
        action = mock.Mock()
        action.action = consts.CLUSTER_SCALE_IN
        action.context = self.context
        action.data = {}
        action.inputs = {'count': 2}

        cluster = mock.Mock(nodes=[mock.Mock()])
        mock_load.return_value = cluster
        mock_select.return_value = ['NODE_ID']

        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)

        mock_load.assert_called_once_with(action.context,
                                          dbcluster=None, cluster_id='FAKE_ID')
        mock_update.assert_called_once_with(action, ['NODE_ID'])
        # the following was invoked with 1 because the input count is
        # greater than the cluster size
        mock_select.assert_called_once_with(cluster.nodes, 1, True)

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'nodes_by_age')
    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_scale_in_without_count(self, mock_load, mock_select,
                                           mock_update):
        action = mock.Mock()
        action.context = self.context
        action.action = consts.CLUSTER_SCALE_IN
        action.data = {}
        action.inputs = {}

        cluster = mock.Mock(nodes=[mock.Mock()])
        mock_load.return_value = cluster
        mock_select.return_value = ['NODE_ID']

        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)

        mock_load.assert_called_once_with(action.context,
                                          dbcluster=None, cluster_id='FAKE_ID')
        mock_update.assert_called_once_with(action, ['NODE_ID'])
        # the following was invoked with 1 because the input count is
        # not specified so 1 becomes the default
        mock_select.assert_called_once_with(cluster.nodes, 1, True)

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'parse_resize_params')
    @mock.patch.object(co.Cluster, 'get')
    def test_pre_op_resize_failed_parse(self, mock_get, mock_parse,
                                        mock_update):
        action = mock.Mock()
        action.context = self.context
        action.action = consts.CLUSTER_RESIZE
        action.inputs = {}
        action.data = {}

        db_cluster = mock.Mock()
        mock_get.return_value = db_cluster
        mock_parse.return_value = 'ERROR', 'Failed parsing.'

        policy = dp.DeletionPolicy('test-policy', self.spec)

        policy.pre_op('FAKE_ID', action)

        self.assertEqual('ERROR', action.data['status'])
        self.assertEqual('Failed parsing.', action.data['reason'])
        mock_get.assert_called_once_with(action.context, 'FAKE_ID',
                                         project_safe=True)
        mock_parse.assert_called_once_with(action, db_cluster)
        self.assertEqual(0, mock_update.call_count)

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'parse_resize_params')
    @mock.patch.object(co.Cluster, 'get')
    def test_pre_op_resize_not_deletion(self, mock_get, mock_parse,
                                        mock_update):
        def fake_parse(a, c):
            a.data = {}
            return 'OK', 'cool'

        action = mock.Mock()
        action.context = self.context
        action.action = consts.CLUSTER_RESIZE
        action.inputs = {}

        db_cluster = mock.Mock()
        mock_get.return_value = db_cluster
        mock_parse.side_effect = fake_parse

        policy = dp.DeletionPolicy('test-policy', self.spec)
        # a simulation of non-deletion RESZIE
        action.data = {}

        policy.pre_op('FAKE_ID', action)

        mock_get.assert_called_once_with(action.context, 'FAKE_ID',
                                         project_safe=True)
        mock_parse.assert_called_once_with(action, db_cluster)
        self.assertEqual(0, mock_update.call_count)

    @mock.patch.object(su, 'parse_resize_params')
    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'nodes_by_age')
    @mock.patch.object(cm.Cluster, 'load')
    @mock.patch.object(co.Cluster, 'get')
    def test_pre_op_resize_with_count(self, mock_get, mock_load,
                                      mock_select, mock_update,
                                      mock_parse):
        def fake_parse(a, obj):
            a.data = {
                'deletion': {
                    'count': 2
                }
            }
            return 'OK', 'cool'

        action = mock.Mock()
        action.context = self.context
        action.action = consts.CLUSTER_RESIZE
        action.inputs = {}
        action.data = {}

        db_cluster = mock.Mock()
        mock_get.return_value = db_cluster
        mock_parse.side_effect = fake_parse

        cluster = mock.Mock(nodes=[mock.Mock(), mock.Mock()])
        mock_load.return_value = cluster

        mock_select.return_value = ['NID']

        policy = dp.DeletionPolicy('test-policy', self.spec)

        # mock_parse = self.patchobject(su, 'parse_resize_params',
        #                              side_effect=fake_parse)

        policy.pre_op('FAKE_ID', action)

        mock_get.assert_called_once_with(action.context, 'FAKE_ID',
                                         project_safe=True)
        mock_parse.assert_called_once_with(action, db_cluster)
        mock_load.assert_called_once_with(action.context,
                                          dbcluster=db_cluster,
                                          cluster_id='FAKE_ID')
        mock_update.assert_called_once_with(action, ['NID'])

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'nodes_by_random')
    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_do_random(self, mock_load, mock_select, mock_update):
        action = mock.Mock()
        action.context = self.context
        action.inputs = {}
        action.data = {'deletion': {'count': 2}}

        cluster = mock.Mock(nodes=['a', 'b', 'c'])
        mock_select.return_value = ['NODE1', 'NODE2']
        mock_load.return_value = cluster

        self.spec['properties']['criteria'] = 'RANDOM'
        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)

        mock_load.assert_called_once_with(action.context,
                                          dbcluster=None, cluster_id='FAKE_ID')
        mock_select.assert_called_once_with(cluster.nodes, 2)
        mock_update.assert_called_once_with(action, ['NODE1', 'NODE2'])

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'nodes_by_profile_age')
    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_do_oldest_profile(self, mock_load, mock_select,
                                      mock_update):
        action = mock.Mock()
        action.context = self.context
        action.inputs = {}
        action.data = {'deletion': {'count': 2}}

        mock_select.return_value = ['NODE1', 'NODE2']

        cluster = mock.Mock(nodes=['a', 'b', 'c'])
        mock_load.return_value = cluster

        self.spec['properties']['criteria'] = 'OLDEST_PROFILE_FIRST'
        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)

        mock_load.assert_called_once_with(action.context,
                                          dbcluster=None, cluster_id='FAKE_ID')
        mock_select.assert_called_once_with(cluster.nodes, 2)
        mock_update.assert_called_once_with(action, ['NODE1', 'NODE2'])

    @mock.patch.object(dp.DeletionPolicy, '_update_action')
    @mock.patch.object(su, 'nodes_by_age')
    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_do_oldest_first(self, mock_load, mock_select, mock_update):
        action = mock.Mock()
        action.context = self.context
        action.inputs = {}
        action.data = {'deletion': {'count': 2}}

        cluster = mock.Mock(nodes=['a', 'b', 'c'])
        mock_select.return_value = ['NODE1', 'NODE2']
        mock_load.return_value = cluster

        self.spec['properties']['criteria'] = 'OLDEST_FIRST'
        policy = dp.DeletionPolicy('test-policy', self.spec)
        policy.pre_op('FAKE_ID', action)

        mock_load.assert_called_once_with(action.context,
                                          dbcluster=None, cluster_id='FAKE_ID')
        mock_select.assert_called_once_with(cluster.nodes, 2, True)
        mock_update.assert_called_once_with(action, ['NODE1', 'NODE2'])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from senlin.api.common import util
from senlin.api.common import wsgi
from senlin.common import consts
from senlin.common.i18n import _
from senlin.common import utils


class ActionData(object):
    """All required data fields for an action."""

    PARAMS = (consts.ACTION_NAME, consts.ACTION_TARGET, consts.ACTION_ACTION)

    def __init__(self, data):
        self.data = data

    def name(self):
        if consts.ACTION_NAME not in self.data:
            raise exc.HTTPBadRequest(_("No action name specified"))
        return self.data[consts.ACTION_NAME]

    def target(self):
        if consts.ACTION_TARGET not in self.data:
            raise exc.HTTPBadRequest(_("No target specified"))
        return self.data[consts.ACTION_TARGET]

    def action(self):
        if consts.ACTION_ACTION not in self.data:
            raise exc.HTTPBadRequest(_("No action specified"))
        return self.data[consts.ACTION_ACTION]

    def params(self):
        data = self.data.items()
        return dict((k, v) for k, v in data if k not in self.PARAMS)


class ActionController(wsgi.Controller):
    """WSGI controller for Actions in Senlin v1 API."""

    # Define request scope (must match what is in policy.json)
    REQUEST_SCOPE = 'actions'

    @util.policy_enforce
    def index(self, req):
        filter_whitelist = {
            consts.ACTION_NAME: 'mixed',
            consts.ACTION_TARGET: 'mixed',
            consts.ACTION_ACTION: 'mixed',
            consts.ACTION_STATUS: 'mixed',
        }
        param_whitelist = {
            consts.PARAM_LIMIT: 'single',
            consts.PARAM_MARKER: 'single',
            consts.PARAM_SORT: 'single',
            consts.PARAM_GLOBAL_PROJECT: 'single',
        }
        for key in req.params.keys():
            if (key not in param_whitelist.keys() and key not in
                    filter_whitelist.keys()):
                raise exc.HTTPBadRequest(_('Invalid parameter %s') % key)
        params = util.get_allowed_params(req.params, param_whitelist)
        filters = util.get_allowed_params(req.params, filter_whitelist)

        key = consts.PARAM_LIMIT
        if key in params:
            params[key] = utils.parse_int_param(key, params[key])

        key = consts.PARAM_GLOBAL_PROJECT
        if key in params:
            global_project = utils.parse_bool_param(key, params[key])
            params.pop(key)
            params['project_safe'] = not global_project

        if not filters:
            filters = None

        actions = self.rpc_client.action_list(req.context, filters=filters,
                                              **params)

        return {'actions': actions}

    @util.policy_enforce
    def create(self, req, body):
        data = ActionData(body)
        result = self.rpc_client.action_create(req.context,
                                               data.name(),
                                               data.target(),
                                               data.action(),
                                               data.params())

        return result

    @util.policy_enforce
    def get(self, req, action_id):
        action = self.rpc_client.action_get(req.context, action_id)
        if not action:
            raise exc.HTTPNotFound()

        return {'action': action}
